% the subtree of an sst(_9_) - for convenience
st(sst(_,_,_,_,_,_,_,_,ST), ST) :- !.

% numNodes(+Tree, -NumNodes) :- -NumNodes is the number of nodes in +Tree.
numNodes([], 0).
numNodes(sst(_,_,_,_,_,_,_,_,ST), Num) :-
	numNodes(ST, Tmp),
	Num is Tmp + 1.
numNodes([sst(_,_,_,_,_,_,_,_,ST)|Ns], Num) :-
	numNodes(ST, Tmp1),
	numNodes(Ns, Tmp2),
	Tmp3 is Tmp1 + Tmp2,
	Num is Tmp3 + 1.

% recover only the marked nodes of an mtree (assuming there are no interrupted paths leading to them)
get_mtree(sst(I,0,S,T,Pi,PIK,L,R,_ST), sst(I,0,S,T,Pi,PIK,L,R,[])).
get_mtree(sst(I,1,S,T,Pi,PIK,L,R,[]), sst(I,1,S,T,Pi,PIK,L,R,[])).
get_mtree(sst(I,1,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,SST)) :-
	get_mtree_list(ST, SST).
get_mtree_list([], []).
get_mtree_list([sst(I,1,S,T,Pi,PIK,L,R,ST)|Ns], [sst(I,1,S,T,Pi,PIK,L,R,SST)|NNs]) :-
	get_mtree_list(ST, SST),
	get_mtree_list(Ns, NNs).
get_mtree_list([sst(_,0,_,_,_,_,_,_,_)|Ns], NNs) :-
	get_mtree_list(Ns, NNs).

% path(+Tree, ?Path) :- ?Path is a linear path in +Tree of unspecified length where +Tree is an sst(_9_).
path(sst(I,M,S,T,Pi,PIK,L,R,_), sst(I,M,S,T,Pi,PIK,L,R,[])).
path(sst(I,M,S,T,Pi,PIK,L,R,ST), sst(I,M,S,T,Pi,PIK,L,R,[Path])) :-
	member(X, ST),
	path(X, Path).
% path_with_subnode(+Tree, ?Path) :- same as path, only with subnodes kept at the leaf
path_with_subnode(sst(I,M,S,T,Pi,PIK,L,R,ST), sst(I,M,S,T,Pi,PIK,L,R,ST)).
path_with_subnode(sst(I,M,S,T,Pi,PIK,L,R,ST), sst(I,M,S,T,Pi,PIK,L,R,[Path])) :-
	member(X, ST),
	path_with_subnode(X, Path).
% as the above variants with the exeption that all nodes belonging to the path are marked '1'
marked_path(sst(I,_,S,T,Pi,PIK,L,R,_), sst(I,1,S,T,Pi,PIK,L,R,[])).
marked_path(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,[Path])) :-
	member(X, ST),
	marked_path(X, Path).
marked_path_with_subnode(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,ST)).
marked_path_with_subnode(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,[Path])) :-
	member(X, ST),
	marked_path_with_subnode(X, Path).
marked_path_with_subnode_and_length(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,ST), 0).
marked_path_with_subnode_and_length(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,[Path]), Len) :-
	member(X, ST),
	marked_path_with_subnode_and_length(X, Path, Len1),
	Len is Len1 + 1.
% marked_path_with_subnode_and_length/4: also gives the path without subnodes - for testing
marked_path_with_subnode_and_length(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,ST), 0, sst(I,1,S,T,Pi,PIK,L,R,[])).
marked_path_with_subnode_and_length(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,[Path]), Len, sst(I,1,S,T,Pi,PIK,L,R,[Path])) :-
	member(X, ST),
	marked_path_with_subnode_and_length(X, Path, Len1, PathNSn),
	Len is Len1 + 1.

% subnode(+?Subnode, +Tree) :- succeeds if +?Subnode is a subnode in +Tree.
subnode(sst(I,M,S,T,Pi,PIK,L,R,ST), sst(I,M,S,T,Pi,PIK,L,R,ST)).
subnode(SN, sst(_,_,_,_,_,_,_,_,ST)) :-
	member(X, ST),
	subnode(SN, X).
% subnode(?Subnode, +Index, +Tree) :- subnode by node-index.
subnode(X, Index, T) :-
	subnode(X, T),
	X = sst(Index,_,_,_,_,_,_,_,_).

% leafs(+Tree, ?ListOfLeafs)
leafs(sst(I,M,S,T,Pi,PIK,L,R,[]), [sst(I,M,S,T,Pi,PIK,L,R,[])]).
leafs(sst(_,_,_,_,_,_,_,_,STL), L) :-
	STL \== [],
	leafs_l(STL, L).
leafs_l([],[]).
leafs_l([T|Ts], LLL) :-
	leafs(T,L),
	leafs_l(Ts, LL),
	append(L,LL,LLL).

% set_mtree(+MTree, ?OneMTree) :- OneMtree has sst(_,1,...) for every node.
set_mtree(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,SST)) :-
	set_mtree_list(ST, SST).
set_mtree_list([], []).
set_mtree_list([T|Ts], [TT|TTs]) :-
	set_mtree(T, TT),
	set_mtree_list(Ts, TTs).
% reset_mtree(+MTree, ?NulledMTree) :- NulledMtree has sst(0,...) for every node.
reset_mtree(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,0,S,T,Pi,PIK,L,R,SST)) :-
	reset_mtree_list(ST, SST).
reset_mtree_list([], []).
reset_mtree_list([T|Ts], [TT|TTs]) :-
	reset_mtree(T, TT),
	reset_mtree_list(Ts, TTs).

% replace_node_in_mtree(+OldNode, +OldTree, +NewNode, ?NewTree) :- NewTree is obtained from OldTree by replacing the first occurrence of OldNode in OldTree by NewNode.
replace_node_in_mtree(Node, Node, NewNode, NewNode).
replace_node_in_mtree(Node, sst(I,M,S,T,Pi,PIK,L,R,ST), NewNode, Result) :-
	member(X, ST),
	replace_node_in_mtree(Node, X, NewNode, NewX),
	rle(X, NewX, ST, SST),										% test.pl
	Result = sst(I,M,S,T,Pi,PIK,L,R,SST).

% no marked children (in tree-list)
no_marked_ch([]) :- !.
no_marked_ch([sst(_,0,_,_,_,_,_,_,_)|R]) :-
	no_marked_ch(R).

% marked children (in tree-list)
marked_ch([sst(_,1,_,_,_,_,_,_,_)|_]) :-
	true.
marked_ch([sst(_,0,_,_,_,_,_,_,_)|Xs]) :-
	marked_ch(Xs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%  From a given sst(_9_) create a new structure that contains real prolog variable instances instead of var/1 and var/2 terms %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get_variable_tree(+Tree, -SubstList, -VarTree) :- -VarTree is +Tree with every term (var(_) or var(_,_) replaced by a prolog variable,
%													where each instance of var(X) or var(X,Y) gets replaced by an instance of the same prolog var.
get_variable_tree(Tree, SubstList, VarTree) :-
	vars(Tree, Vars),
	assign_vv_t(Vars, SubstList),
	assign_vars_term(Tree, SubstList, VarTree).
%%%%%%%%%%%%%%%%%%% helper predicates %%%%%%%%%%%%%%%%%%%
assign_vv_t([], []).
assign_vv_t([V|Vs], [(V,P)|VPs]) :-		% will throw a warning on compilation: ignore
	assign_vv_t(Vs, VPs).
% var-case
assign_vars_term(var(X), SubstList, Var) :-
	member((var(X), Var), SubstList), !.
assign_vars_term(var(X,K), SubstList, Var) :-
	member((var(X,K), Var), SubstList), !.
% atoms & vars
assign_vars_term(T, _SL, T) :-
	atomic(T), !.
assign_vars_term(T, _SL, T) :-
	var(T), !.
% compound terms
assign_vars_term(Term, SubstList, NewTerm) :-
	Term =.. [Head|Tail],
	assign_vars_term_list(Tail, SubstList, NewTail),
	NewTerm =.. [Head|NewTail].
assign_vars_term_list([], _, []).
assign_vars_term_list([T|Ts], SL, [NT|NTs]) :-
	assign_vars_term(T, SL, NT),
	assign_vars_term_list(Ts, SL, NTs).

% retrieve variables in a sst(_9_)
vars(sst(_,_,_,_,_,_,L,R,[]), Vars) :-
	varsterm(L, VL),
	varsterm(R, VR),
	append(VL,VR,Vars), !.
vars(sst(_,_,_,_,_,_,L,R,ST), UVars) :-
	ST \== [],
	varsterm(L, VL),
	varsterm(R, VR),
	varsl(ST, VST),
	append(VL,VR,Tmp),
	append(VST, Tmp, Vars),
	remove_duplicates(Vars, UVars), !.
varsl([],[]) :-!.
varsl([T|Ts], Vars) :-
	vars(T, VT),
	varsl(Ts, VTs),
	append(VT, VTs, Vars).
varsterm(X, []) :-
	atomic(X), !.
varsterm(var(X,R), [var(X,R)|Rest]) :-
	varsterm(X,Rest).
varsterm(var(X), [var(X)|Rest]) :-
	varsterm(X, Rest).
varsterm([X|Y], Vars) :-
	varsterm(X, XV),
	varsterm(Y, YV),
	append(XV, YV, Vars).
varsterm(hash(X), Rest) :-
	varsterm(X, Rest).
% sc
varsterm(sc(_,_,X), Vars) :-
	varsterm(X, Vars).
% sym_enc, asym_enc, sig
varsterm(X, Vars) :-
	compound(X),
	X =.. [_Functor|Ts],
	varsterm(Ts, Vars).
% to be complete ...
varsterm([], []) :- !.