% solution_exists(+ConstraintSystem, ?Result) :- invoke the constraint solver on +ConstraintSystem, also print one dot to stdout.
solution_exists(CC,Res) :-
	nb_getval(csolve_calls, C),
	C1 is C + 1,
	nb_setval(csolve_calls, C1),
	flush,
	nb_getval(verbose, V),
	(V=1 -> write('.'); true),
	reverse(CC, RCC),
	solve(RCC,Res).
solution_exists(_,_) :-
	fail.

% extend_minimal(+Tree, ?MinimalTree) :-	?MinimalTree consists of +Tree, extended by all reachable epsilon- and sc-nodes;
%											where reachable means that a path of marked nodes starting at the root exists.
extend_minimal(sst(I,0,S,T,Pi,PIK,L,R,ST), sst(I,0,S,T,Pi,PIK,L,R,ST)) :-
	T \== root,
	T \== epsilon,
	T \== sc, !.
extend_minimal(sst(I,_,S,T,Pi,PIK,L,R,ST), sst(I,1,S,T,Pi,PIK,L,R,SST)) :-
	extend_minimal_list(ST, SST).
extend_minimal_list([], []).
extend_minimal_list([N|Ns], [NN|NNs]) :-
	extend_minimal(N, NN),
	extend_minimal_list(Ns, NNs).

% extend_deep(+Tree, -ExtendedTree) :-	-ExtendedTree is (if) the extension of Tree; it only adds one node at most;
%										the extension is done at one of the leafs of +Tree.
extend_deep(MT, EMT) :-
	subnode(X, MT),
	X = sst(_,1,_,_,_,_,_,_,XST),
	XST \== [],
	no_marked_ch(XST),
	member(Y0, XST),
	Y0 = sst(YI,0,YS,intruder,YPSi,YPIK,YL,YR,YST),
	Y1 = sst(YI,1,YS,intruder,YPSi,YPIK,YL,YR,YST),
	replace_node_in_mtree(Y0, MT, Y1, EMT).


%% Step A.4.2.b from the algorithm %%%%%%%%%
% extend_wide(+Tree, -ExtendedTree) :-	-ExtendedTree is obtained from +Tree by extending all nodes which result from an
%										intruder transition and which match the LHS of an already existing sibling node in +Tree.
extend_wide(sst(I,M,S,T,P,PIK,L,R,[]), sst(I,M,S,T,P,PIK,L,R,[])) :- !.
extend_wide(sst(I,M,S,T,P,PIK,L,R,ST), EFMT) :-
	member(X, ST),
	X = sst(_,1,_,intruder,_,_,XL,_,_),
	findall(Y, (member(Y, ST),
				Y=sst(_,0,_,intruder,_,_,YL,_,_),
				%XL == YL), %NEW
				match(XL, YL)), % match prolog vars also, without instantiating
				YList),
	set_all_nodes_one(ST, YList, SST),
	extend_wide(X, NX),
	(X\==NX,!,replace_node_in_mtree(X, sst(I,M,S,T,P,PIK,L,R,SST), NX, EFMT); EFMT=sst(I,M,S,T,P,PIK,L,R,SST)),!.
extend_wide(sst(I,M,S,T,P,PIK,L,R,ST), sst(I,M,S,T,P,PIK,L,R,SST)) :-
	extend_wide_list(ST, SST).
extend_wide_list([], []).
extend_wide_list([N|Ns], [NN|NNs]) :-
	extend_wide(N, NN),
	extend_wide_list(Ns, NNs).

% set_all_nodes_one(+NodeList, +ZeroNodeList, ?NewNodeList) :-  all items that are member from +ZeroNodeList and +NodeList
%																are being marked and form ?NewNodeList.
% needed by extend_wide/2
set_all_nodes_one(NL, [], NL).
set_all_nodes_one(NL, [N|Ns], NNNL) :-
	set_node_one(NL, N, NNL),
	set_all_nodes_one(NNL, Ns, NNNL).
% set_node_one(+NodeList, +ZeroNode, ?NewNodeList) :- assuming that member(+ZeroNode, +NodeList) is true.
set_node_one([sst(I,0,S,T,P,PIK,L,R,ST)|Xs], sst(I,0,S,T,P,PIK,L,R,ST), [sst(I,1,S,T,P,PIK,L,R,ST)|Xs]) :- !.
set_node_one([sst(XI,XM,XS,XT,XP,XPIK,XL,XR,XST)|Xs], sst(I,0,S,T,P,PIK,L,R,ST), [sst(XI,XM,XS,XT,XP,XPIK,XL,XR,XST)|Ys]) :-
	sst(XI,XM,XS,XT,XP,XPIK,XL,XR,XST) \== sst(I,0,S,T,P,PIK,L,R,ST),
	set_node_one(Xs, sst(I,0,S,T,P,PIK,L,R,ST), Ys).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% syntax for applying some unique atoms to the variables in a simple CS as returned by the solver 		%
% this should be used locally in a not-negative check of a temporary solution							%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% numbervars(SimpleCS,0,_,[functor_name(uvar)])
check_minimal(T, Pos, Neg, Final) :-
	extend_minimal(T, MT),
	%(nb_getval(debug,1) -> (write('\nminimal'), own_debug(MT))),
	check_only_branch_constraints(MT), !,
	check_constraint_system(MT, Pos, SimpleCS),
	check_not_negative(MT, Pos, Neg, SimpleCS, Final).
check_minimal(T, Pos, Neg, Final) :-
	extend_minimal(T, MT),
	check_deep(MT, Pos, Neg, Final).

% coming from min-check
check_not_negative(T, Pos, Neg, SimpleCS, Final) :-
	%(nb_getval(debug,1) -> (write('\ncnn/5 '), own_debug(T))),
	retractall(tmp(_,_,_)),
	assert(tmp(T, Neg, SimpleCS)),
	not((numbervars(SimpleCS,0,_,[functor_name(uatom)]), any_is_derivable(T, Neg, SimpleCS))), !,			% bad cut ?
	check_postproc(T, Pos, Neg, Final).
% going into min-check
check_not_negative(T, Pos, Neg, Final) :-
	%(nb_getval(debug,1) -> (write('\ncnn/4 '), own_debug(T))),
	not(any_is_derivable(T, Neg, [])), !,
	check_minimal(T, Pos, Neg, Final).

check_deep(T, Pos, Neg, Final) :-
	extend_deep(T, ET), (T\==ET; fail),
	%(nb_getval(debug,1) -> (write('\ndeep'), own_debug(ET))),
	check_not_negative(ET, Pos, Neg, Final).

check_postproc(T, _, _, T) :-
	extend_wide(T, T), !,
	finish(T).
check_postproc(T, Pos, Neg, Final) :-
	extend_wide(T, ET),
	T\==ET,
	%(nb_getval(debug,1) -> (write('\npp'), own_debug(ET))),
	check_not_negative(ET, Pos, Neg, Final).

finish(Final) :- !,
	nb_getval(csolve_calls, CALLS),
	nb_getval(verbose, V),
	(V=1 -> writef("  constraint solver called %w times\n", [CALLS]); true).
	%retractall(successfultree(_)),
	%assert(successfultree(Final)).
	%writeln('\nfinished.').

% true if any of Neg is derivable from Tree
% TODO: assuming that the variable substitution done by numbervars to SimpleCS still applies here
%any_is_derivable(_, [], _) :- !, fail.
any_is_derivable(T, Neg, SimpleCS) :-
	atoms_into_leafs(T, SimpleCS, ET),
	check_derivable(ET, Neg).

check_derivable(_, []) :- !, fail.
check_derivable(T, [N|_]) :-
	check_leaf_constraints(T, [N]), !. % one is enough
%	check_constraint_system(T, [N], _), !.
check_derivable(T, [_|Ns]) :-
	check_derivable(T, Ns).

% all atoms, all leafs
atoms_into_leafs(T, [], T).
atoms_into_leafs(T, [Constr|Rest], EET) :-
	single_atom_into_leafs(T, Constr, ET),
	atoms_into_leafs(ET, Rest, EET).
% single atom, all leafs
single_atom_into_leafs(T, [Atom,_], ET) :-
	leafs(T, L),
	add_atom_to_leafs(T, Atom, L, ET).
% single atom, single leaf
add_atom_to_leafs(T, _, [], T).
add_atom_to_leafs(T, Atom, [L|Ls], EET) :-
	L = sst(I,M,state(PL,IK,SC),Tr,P,PIK,LHS,RHS,ST),
	append(IK,[Atom],NewIK),
	NewL = sst(I,M,state(PL,NewIK,SC),Tr,P,PIK,LHS,RHS,ST),
	replace_node_in_mtree(L, T, NewL, ET),
	add_atom_to_leafs(ET, Atom, Ls, EET).



