% epsilon transition
sym_transition(U, U_, epsilon) :-
	U = state(PList, IK, SC),
	member(Pi, PList),
	Pi = p(PiChilList),
 	member(PSi, PiChilList),
	PSi = ps(rule(epsilon, RHS),VPi),

	replace_in_list(Pi, VPi, PList, NewPList),
	adj_intrKn_sc(IK, SC, RHS, NewSC, NewIK),
	nb_getval(nc, NC),
	NNC is NC + 1,
	nb_setval(nc, NNC),
	U_ = sst(NNC, 0, state(NewPList, NewIK, NewSC),
					epsilon, PSi, [], epsilon, RHS, []).

% secure channel transition
sym_transition(U, U_, sc) :-
	U = state(PList, IK, SC),
	member(Pi, PList),
	Pi = p(PiChilList),

	member(PSi, PiChilList),
	PSi = ps(rule(LHS, RHS),VPi),
	member(R_, SC),
	match(LHS, R_),
	LHS = sc(_,_,MSG),
	variables(MSG, VarsInLHS),

	nb_getval(nc, NC),
	NNC is NC + 1,
	nb_setval(nc, NNC),

	rename_vars_in_principal(VPi, VarsInLHS, NNC, NewVPi),
	replace_in_list(Pi, NewVPi, PList, NewPList),
	rename_vars_in_term(RHS, VarsInLHS, NNC, NewRHS),
	rename_vars_in_term(LHS, VarsInLHS, NNC, LHS_Ren),

	adj_intrKn_sc(IK, SC, NewRHS, TmpSC, NewIK),
	adj_sc(TmpSC, R_, NewSC),

	U_ = sst(NNC, 0, state(NewPList, NewIK, NewSC),
					sc, PSi, [], LHS_Ren, NewRHS, []).

% intruder transition
sym_transition(U,U_, intruder) :-
	U = state(PList, IK, SC),
	member(Pi, PList),
	Pi = p(PiChilList),

	member(PSi, PiChilList),
	PSi = ps(rule(LHS, RHS),VPi),
	LHS \= sc(_,_,_),
	LHS \= epsilon,
%assert(lhs(LHS)), abort,
	variables(LHS, VarsInLHS),

	nb_getval(nc, NC),
	NNC is NC + 1,
	nb_setval(nc, NNC),

	rename_vars_in_principal(VPi, VarsInLHS, NNC, NewVPi),
	rename_vars_in_term(RHS, VarsInLHS, NNC, NewRHS),
	rename_vars_in_term(LHS, VarsInLHS, NNC, LHS_Ren),

	adj_intrKn_sc(IK, SC, NewRHS, NewSC, NewIK),
	replace_in_list(Pi, NewVPi, PList, NewPList),

	U_ = sst(NNC, 0, state(NewPList, NewIK, NewSC),
					intruder, PSi, [], LHS_Ren, NewRHS, []).

% take a specific intruder transition, used by step A 4.2 of the algorithm
% sym_transition(+InputState, +PrincipalToTake, ?OutputSSTNode) :- Type of transition 'intruder' is assumed here.
fixed_principal_sym_transition(U, PSi, U_) :-
	U = state(PList, IK, SC),
	member(Pi, PList),
	Pi = p(PiChilList),

	member(PSi, PiChilList),
	PSi = ps(rule(LHS, RHS),VPi),
	LHS \= sc(_,_,_),
	LHS \= epsilon,
	variables(LHS, VarsInLHS),

	nb_getval(nc, NC),
	NNC is NC + 1,
	nb_setval(nc, NNC),

	rename_vars_in_principal(VPi, VarsInLHS, NNC, NewVPi),
	rename_vars_in_term(RHS, VarsInLHS, NNC, NewRHS),
	rename_vars_in_term(LHS, VarsInLHS, NNC, LHS_Ren),

	adj_intrKn_sc(IK, SC, NewRHS, NewSC, NewIK),
	replace_in_list(Pi, NewVPi, PList, NewPList),

	U_ = sst(NNC, 0, state(NewPList, NewIK, NewSC),
					intruder, PSi, [], LHS_Ren, RHS, []).


/* ** step(a) - rename variables in the principal subtree ****
 *************************************************************/
rename_vars_in_principal(P, [], _ReplIndex, P).
rename_vars_in_principal(P, [V|Vs], ReplIndex, ResP) :-
	replace_principal(P, V, ReplIndex, TmpP),
	rename_vars_in_principal(TmpP, Vs, ReplIndex, ResP).

replace_principal(p([]), _V, _, p([])).

replace_principal(p([S|Ss]), V, ReplIndex, p([S_|Ss_])) :-
	replace_successor(S, V, ReplIndex, S_),
	replace_principal(p(Ss), V, ReplIndex, p(Ss_)).

replace_successor(ps(rule(LHS, RHS),P), V, ReplIndex, ps(rule(NewLHS, NewRHS),NewP)) :-
	rename_single(LHS, V, ReplIndex, NewLHS),
	rename_single(RHS, V, ReplIndex, NewRHS),
	replace_principal(P, V, ReplIndex, NewP).

% (b,c)
% case: RHS is a secure channel term
adj_intrKn_sc(I, SC, sc(N,N_,T), SC_, I_) :-
	append(I, [T], I_),
	append(SC, [sc(N,N_,T)], SC_),	% add to secure channel
	!.
% case: RHS is _not_ a secure channel term
adj_intrKn_sc(I, SC, T, SC, I_) :-
	append(I, [T], I_), !.

% adj_sc(+SC, +Term, ?NewSC) :- ?NewSC is +SC \ +Term.
adj_sc(SC, sc(A,B,C), SC_) :-
	append([sc(A,B,C)], SC_, SC), !.
adj_sc(SC, _, SC) :- !.

