%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% MAIN																																	%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generating the full tree
full_tree(INIT, T22p) :-
	nb_getval(verbose, V),
	(V=1 -> writef("\nBuilding full tree from input state ... "); true),
	reset,
	build_tree(INIT, T),
	a22([T],[T22]),
	pik(T22, [], T22p),
	(V=1 -> writef(" ok\n\n"); true).

% build_tree(+InitialNode, -Tree) :- -Tree is successively build from +InitialNode by applying symbolic transitions
build_tree(sst(I,M,S,T,Pi,PIK,L,R,_), sst(I,M,S,T,Pi,PIK,L,R,ST)) :-
	findall(Node, sym_transition(S,Node,_), NodeList),
	build_tree_from_node_list(NodeList, ST).
build_tree_from_node_list([],[]).
build_tree_from_node_list([N|Ns],R) :-
	build_tree_from_node_list(Ns, NNs),
	build_tree(N, NN),
	append([NN], NNs, R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% HELPER																																	%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% recursively append a specific term to the IK of each sst-Node in a symbolic tree
% append_to_IK(+Node, +Term, ?NewNode)
append_to_IK([], _, []).
% case: +Term is already a member: IK stays the same
append_to_IK([sst(I,M,state(P,IK,SC),T,Pi,PIK,L,R,ST)|Ns], Term, [sst(I,M,state(P,IK,SC),T,Pi,PIK,L,R,SST)|NNs]) :-
	member(Term, IK),
	append_to_IK(ST, Term, SST),
	append_to_IK(Ns, Term, NNs).
% case: the rest: IK -> IIk
append_to_IK([sst(I,M,state(P,IK,SC),T,Pi,PIK,L,R,ST)|Ns], Term, [sst(I,M,state(P,IIK,SC),T,Pi,PIK,L,R,SST)|NNs]) :-
	append([Term], IK, IIK),
	append_to_IK(ST, Term, SST),
	append_to_IK(Ns, Term, NNs).

% step A-2.2: "Encoding secure channel communication'
%% a22(+SymTree, ?NSymTree)
a22([], []).
% case: msg written into sc
a22([sst(I,M,S,T,Pi,PIK,L,sc(N,N_,R),ST)|Ns], [NewNode|NNs]) :-
	a22_encrypt(sst(I,M,S,T,Pi,PIK,L,sc(N,N_,R),ST),NewNode),
	a22(Ns, NNs), !.
% case: any other msg
a22([sst(I,M,state(P,IK,SC),T,Pi,PIK,L,R,ST)|Ns], [sst(I,M,state(P,IK,SC),T,Pi,PIK,L,R,SST)|NNs]) :-
	a22(ST, SST),
	a22(Ns, NNs).
a22_encrypt(sst(I,M,state(P,IK,SC),T,Pi,PIK,LHS,sc(Addr,Addr_,R),ST), sst(I,M,state(P,IIK,SC),T,Pi,PIK,LHS,sc(Addr,Addr_,R),SSSST)) :-
	% keygen:
	nb_getval(kc, KC),									% key counter
	atom_concat(sc_key_, KC, ThisKey),
	% append to intruder knowledge lists
	% NEW: replaced sym_enc(Val, Key) by Val>Key %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%append([sym_enc(R,ThisKey)],IK,IIK),						% IK is N's intruder kn.
	%append_to_IK(ST, sym_enc(R,ThisKey), SST),				% tools.pl
	append([otk(R,ThisKey)],IK,IIK),
	append_to_IK(ST, otk(R,ThisKey), SST),
	% NEW END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	KCC is KC + 1,
	nb_setval(kc, KCC),
	% replace LHS of successors by encrypted R' if LSH!=sc(n,n',R)
	a22_check_reachables(SST, sc(Addr,Addr_,R), ThisKey, SSST), %NEW changed by above 'new'
	% check subtree for further sc-msgs
	a22(SSST, SSSST).
a22_check_reachables([], _, _, []).
% case: edge label = [_,_,sc(n,n',R),sc]: ignore this edge
a22_check_reachables([sst(I,M,S,sc,Pi,PIK,SC_Term,RHS,ST)|Ns], SC_Term, KC, [sst(I,M,S,sc,Pi,PIK,SC_Term,RHS,SST)|NNs]) :-
	a22_check_reachables(Ns, SC_Term, KC, NNs),
	a22_check_reachables(ST, SC_Term, KC, SST),
	!.
% case: no sc-msg
a22_check_reachables([sst(I,M,S,T,Pi,PIK,L,R,ST)|Ns], SC_Term, KC, [sst(I,M,S,T,Pi,PIK,L,R,SST)|NNs]) :-
	T \== sc,
	a22_check_reachables(ST, SC_Term, KC, SST),
	a22_check_reachables(Ns, SC_Term, KC, NNs).
% case: sc-message of the form [_,_,sc(n,n',R'),sc]
% TODO: replace the label of this edge with sym_enc(R,KC) ?
a22_check_reachables([sst(I,M,S,sc,Pi,PIK,sc(_,_,R),RHS,ST)|Ns], SC_Term, KC, [sst(I,M,S,sc,Pi,PIK,NewLHS,RHS,SST)|NNs]) :-
	%NewLHS = sym_enc(R,KC), % NEW: changed by 'new' above: KC is now 'otk(RHS,OneTimeKey)'
	NewLHS = otk(R,KC),
	a22_check_reachables(ST, SC_Term, KC, SST),
	a22_check_reachables(Ns, SC_Term, KC, NNs).

% PROPAGAte IK to list of nodes
% case: leaf node:
% introducing a new type of sst()-node by adding the leaf's old IK right behind the principal information
pik(sst(I,M,state(PL,IK,SC),Ty,Pi,_PIK,LHS,RHS,[]), NIK, sst(I,M,state(PL,IK,SC),Ty,Pi,NIK,LHS,RHS,[])).
% case root node:
pik(sst(I,M,state(PL,IK,SC),root,Pi,_PIK,LHS,RHS,Ch), _,sst(I,M,state(PL,IK,SC),root,Pi,[],LHS,RHS,NewCh)) :- !,
	pik_list(Ch, IK, NewCh).
pik(sst(I,M,state(PL,IK,SC),Ty,Pi,_PIK,LHS,RHS,Ch), NIK, sst(I,M,state(PL,IK,SC),Ty,Pi,NIK,LHS,RHS,NewCh)) :-
	Ch \== [],
	pik_list(Ch, IK, NewCh).
pik_list([], _, []).
pik_list([C|Cs], IK, [NC|NCs]) :-
	pik(C, IK, NC),
	pik_list(Cs, IK, NCs).